@@ -34,16 +34,12 @@
 #endif
 #include <sys/types.h>
 
-#ifdef G_OS_WIN32
-int dont_define_prefix = FALSE;
+gboolean parse_strict = TRUE;
+gboolean define_prefix = ENABLE_DEFINE_PREFIX;
 char *prefix_variable = "prefix";
-int msvc_syntax = FALSE;
-#endif
 
 #ifdef G_OS_WIN32
-#ifndef G_IS_DIR_SEPARATOR
-#define G_IS_DIR_SEPARATOR(c) ((c) == G_DIR_SEPARATOR || (c) == '/')
-#endif
+gboolean msvc_syntax = FALSE;
 #endif
 
 /**
@@ -204,8 +200,8 @@ trim_and_sub (Package *pkg, const char *str, const char *path)
             {
               verbose_error ("Variable '%s' not defined in '%s'\n",
                              varname, path);
-              
-              exit (1);
+              if (parse_strict)
+                exit (1);
             }
 
           g_free (varname);
@@ -234,8 +230,10 @@ parse_name (Package *pkg, const char *str, const char *path)
   if (pkg->name)
     {
       verbose_error ("Name field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
   
   pkg->name = trim_and_sub (pkg, str, path);
@@ -247,8 +245,10 @@ parse_version (Package *pkg, const char *str, const char *path)
   if (pkg->version)
     {
       verbose_error ("Version field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
   
   pkg->version = trim_and_sub (pkg, str, path);
@@ -260,8 +260,10 @@ parse_description (Package *pkg, const char *str, const char *path)
   if (pkg->description)
     {
       verbose_error ("Description field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
   
   pkg->description = trim_and_sub (pkg, str, path);
@@ -290,10 +292,10 @@ typedef enum
 
 #define PARSE_SPEW 0
 
-static GSList*
+static GList *
 split_module_list (const char *str, const char *path)
 {
-  GSList *retval = NULL;
+  GList *retval = NULL;
   const char *p;
   const char *start;
   ModuleSplitState state = OUTSIDE_MODULE;
@@ -374,7 +376,7 @@ split_module_list (const char *str, const char *path)
         {
           /* We left a module */
           char *module = g_strndup (start, p - start);
-          retval = g_slist_prepend (retval, module);
+          retval = g_list_prepend (retval, module);
 
 #if PARSE_SPEW
           fprintf (stderr, "found module: '%s'\n", module);
@@ -392,7 +394,7 @@ split_module_list (const char *str, const char *path)
     {
       /* get the last module */
       char *module = g_strndup (start, p - start);
-      retval = g_slist_prepend (retval, module);
+      retval = g_list_prepend (retval, module);
 
 #if PARSE_SPEW
       fprintf (stderr, "found module: '%s'\n", module);
@@ -400,22 +402,21 @@ split_module_list (const char *str, const char *path)
       
     }
   
-  retval = g_slist_reverse (retval);
+  retval = g_list_reverse (retval);
 
   return retval;
 }
 
-GSList*
+GList *
 parse_module_list (Package *pkg, const char *str, const char *path)
 {
-  GSList *split;
-  GSList *iter;
-  GSList *retval = NULL;
+  GList *split;
+  GList *iter;
+  GList *retval = NULL;
 
   split = split_module_list (str, path);
-  
-  iter = split;
-  while (iter != NULL)
+
+  for (iter = split; iter != NULL; iter = g_list_next (iter))
     {
       RequiredVersion *ver;
       char *p;
@@ -426,7 +427,7 @@ parse_module_list (Package *pkg, const char *str, const char *path)
       ver = g_new0 (RequiredVersion, 1);
       ver->comparison = ALWAYS_MATCH;
       ver->owner = pkg;
-      retval = g_slist_prepend (retval, ver);
+      retval = g_list_prepend (retval, ver);
       
       while (*p && MODULE_SEPARATOR (*p))
         ++p;
@@ -445,8 +446,10 @@ parse_module_list (Package *pkg, const char *str, const char *path)
       if (*start == '\0')
         {
           verbose_error ("Empty package name in Requires or Conflicts in file '%s'\n", path);
-          
-          exit (1);
+          if (parse_strict)
+            exit (1);
+          else
+            continue;
         }
       
       ver->name = g_strdup (start);
@@ -478,9 +481,13 @@ parse_module_list (Package *pkg, const char *str, const char *path)
             ver->comparison = NOT_EQUAL;
           else
             {
-              verbose_error ("Unknown version comparison operator '%s' after package name '%s' in file '%s'\n", start, ver->name, path);
-              
-              exit (1);
+              verbose_error ("Unknown version comparison operator '%s' after "
+                             "package name '%s' in file '%s'\n", start,
+                             ver->name, path);
+              if (parse_strict)
+                exit (1);
+              else
+                continue;
             }
         }
 
@@ -497,9 +504,15 @@ parse_module_list (Package *pkg, const char *str, const char *path)
       
       if (ver->comparison != ALWAYS_MATCH && *start == '\0')
         {
-          verbose_error ("Comparison operator but no version after package name '%s' in file '%s'\n", ver->name, path);
-          
-          exit (1);
+          verbose_error ("Comparison operator but no version after package "
+                         "name '%s' in file '%s'\n", ver->name, path);
+          if (parse_strict)
+            exit (1);
+          else
+            {
+              ver->version = g_strdup ("0");
+              continue;
+            }
         }
 
       if (*start != '\0')
@@ -508,14 +521,12 @@ parse_module_list (Package *pkg, const char *str, const char *path)
         }
 
       g_assert (ver->name);
-      
-      iter = g_slist_next (iter);
     }
 
-  g_slist_foreach (split, (GFunc) g_free, NULL);
-  g_slist_free (split);
+  g_list_foreach (split, (GFunc) g_free, NULL);
+  g_list_free (split);
 
-  retval = g_slist_reverse (retval);
+  retval = g_list_reverse (retval);
 
   return retval;
 }
@@ -523,95 +534,39 @@ parse_module_list (Package *pkg, const char *str, const char *path)
 static void
 parse_requires (Package *pkg, const char *str, const char *path)
 {
-  GSList *parsed;
-  GSList *iter;
   char *trimmed;
-  
+
   if (pkg->requires)
     {
       verbose_error ("Requires field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
 
   trimmed = trim_and_sub (pkg, str, path);
-  parsed = parse_module_list (pkg, trimmed, path);
+  pkg->requires_entries = parse_module_list (pkg, trimmed, path);
   g_free (trimmed);
-  
-  iter = parsed;
-  while (iter != NULL)
-    {
-      Package *req;
-      RequiredVersion *ver = iter->data;
-      
-      req = get_package (ver->name);
-
-      if (req == NULL)
-        {
-          verbose_error ("Package '%s', required by '%s', not found\n",
-                         ver->name, pkg->name ? pkg->name : path);
-          
-          exit (1);
-        }
-
-      if (pkg->required_versions == NULL)
-        pkg->required_versions = g_hash_table_new (g_str_hash, g_str_equal);
-      
-      g_hash_table_insert (pkg->required_versions, ver->name, ver);
-      
-      pkg->requires = g_slist_prepend (pkg->requires, req);
-
-      iter = g_slist_next (iter);
-    }
-
-  g_slist_free (parsed);
 }
 
 static void
 parse_requires_private (Package *pkg, const char *str, const char *path)
 {
-  GSList *parsed;
-  GSList *iter;
   char *trimmed;
-  
+
   if (pkg->requires_private)
     {
       verbose_error ("Requires.private field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
 
   trimmed = trim_and_sub (pkg, str, path);
-  parsed = parse_module_list (pkg, trimmed, path);
+  pkg->requires_private_entries = parse_module_list (pkg, trimmed, path);
   g_free (trimmed);
-  
-  iter = parsed;
-  while (iter != NULL)
-    {
-      Package *req;
-      RequiredVersion *ver = iter->data;
-      
-      req = get_package (ver->name);
-
-      if (req == NULL)
-        {
-          verbose_error ("Package '%s', required by '%s', not found\n",
-                         ver->name, pkg->name ? pkg->name : path);
-          
-          exit (1);
-        }
-
-      if (pkg->required_versions == NULL)
-        pkg->required_versions = g_hash_table_new (g_str_hash, g_str_equal);
-      
-      g_hash_table_insert (pkg->required_versions, ver->name, ver);
-      
-      pkg->requires_private = g_slist_prepend (pkg->requires_private, req);
-
-      iter = g_slist_next (iter);
-    }
-
-  g_slist_free (parsed);
 }
 
 static void
@@ -622,8 +577,10 @@ parse_conflicts (Package *pkg, const char *str, const char *path)
   if (pkg->conflicts)
     {
       verbose_error ("Conflicts field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
 
   trimmed = trim_and_sub (pkg, str, path);
@@ -643,7 +600,8 @@ static char *strdup_escape_shell(const char *s)
 		    (s[0] > '=' && s[0] < '@') ||
 		    (s[0] > 'Z' && s[0] < '^') ||
 		    (s[0] == '`') ||
-		    (s[0] > 'z')) {
+		    (s[0] > 'z' && s[0] < '~') ||
+		    (s[0] > '~')) {
 			r[c] = '\\';
 			c++;
 		}
@@ -675,6 +633,7 @@ static void _do_parse_libs (Package *pkg, int argc, char **argv)
   i = 0;
   while (i < argc)
     {
+      Flag *flag = g_new (Flag, 1);
       char *tmp = trim_string (argv[i]);
       char *arg = strdup_escape_shell(tmp);
       char *p;
@@ -691,9 +650,9 @@ static void _do_parse_libs (Package *pkg, int argc, char **argv)
           while (*p && isspace ((guchar)*p))
             ++p;
 
-          pkg->l_libs = g_slist_prepend (pkg->l_libs,
-                                         g_strconcat (l_flag, p, lib_suffix, NULL));
-
+          flag->type = LIBS_l;
+          flag->arg = g_strconcat (l_flag, p, lib_suffix, NULL);
+          pkg->libs = g_list_prepend (pkg->libs, flag);
         }
       else if (p[0] == '-' &&
                p[1] == 'L')
@@ -701,8 +660,10 @@ static void _do_parse_libs (Package *pkg, int argc, char **argv)
           p += 2;
           while (*p && isspace ((guchar)*p))
             ++p;
-	  pkg->L_libs = g_slist_prepend (pkg->L_libs,
-					 g_strconcat (L_flag, p, NULL));
+
+          flag->type = LIBS_L;
+          flag->arg = g_strconcat (L_flag, p, NULL);
+          pkg->libs = g_list_prepend (pkg->libs, flag);
 	}
       else if (strcmp("-framework",p) == 0 && i+1 < argc)
         {
@@ -713,19 +674,23 @@ static void _do_parse_libs (Package *pkg, int argc, char **argv)
           */
           gchar *framework, *tmp = trim_string (argv[i+1]);
 
-	  framework = strdup_escape_shell(tmp);
-          pkg->other_libs = g_slist_prepend (pkg->other_libs,
-                                             g_strconcat(arg, " ", framework, NULL));
+          framework = strdup_escape_shell(tmp);
+          flag->type = LIBS_OTHER;
+          flag->arg = g_strconcat (arg, " ", framework, NULL);
+          pkg->libs = g_list_prepend (pkg->libs, flag);
           i++;
-          g_free(framework);
-          g_free(tmp);
+          g_free (framework);
+          g_free (tmp);
+        }
+      else if (*arg != '\0')
+        {
+          flag->type = LIBS_OTHER;
+          flag->arg = g_strdup (arg);
+          pkg->libs = g_list_prepend (pkg->libs, flag);
         }
       else
-        {
-          if (*arg != '\0')
-            pkg->other_libs = g_slist_prepend (pkg->other_libs,
-                                               g_strdup (arg));
-        }
+        /* flag wasn't used */
+        g_free (flag);
 
       g_free (arg);
 
@@ -748,8 +713,10 @@ parse_libs (Package *pkg, const char *str, const char *path)
   if (pkg->libs_num > 0)
     {
       verbose_error ("Libs field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
   
   trimmed = trim_and_sub (pkg, str, path);
@@ -759,7 +726,13 @@ parse_libs (Package *pkg, const char *str, const char *path)
     {
       verbose_error ("Couldn't parse Libs field into an argument vector: %s\n",
                      error ? error->message : "unknown");
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        {
+          g_free (trimmed);
+          return;
+        }
     }
 
   _do_parse_libs(pkg, argc, argv);
@@ -792,8 +765,10 @@ parse_libs_private (Package *pkg, const char *str, const char *path)
   if (pkg->libs_private_num > 0)
     {
       verbose_error ("Libs.private field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
   
   trimmed = trim_and_sub (pkg, str, path);
@@ -803,7 +778,13 @@ parse_libs_private (Package *pkg, const char *str, const char *path)
     {
       verbose_error ("Couldn't parse Libs.private field into an argument vector: %s\n",
                      error ? error->message : "unknown");
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        {
+          g_free (trimmed);
+          return;
+        }
     }
 
   _do_parse_libs(pkg, argc, argv);
@@ -825,11 +806,13 @@ parse_cflags (Package *pkg, const char *str, const char *path)
   GError *error = NULL;
   int i;
   
-  if (pkg->I_cflags || pkg->other_cflags)
+  if (pkg->cflags)
     {
       verbose_error ("Cflags field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
   
   trimmed = trim_and_sub (pkg, str, path);
@@ -839,12 +822,19 @@ parse_cflags (Package *pkg, const char *str, const char *path)
     {
       verbose_error ("Couldn't parse Cflags field into an argument vector: %s\n",
                      error ? error->message : "unknown");
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        {
+          g_free (trimmed);
+          return;
+        }
     }
 
   i = 0;
   while (i < argc)
     {
+      Flag *flag = g_new (Flag, 1);
       char *tmp = trim_string (argv[i]);
       char *arg = strdup_escape_shell(tmp);
       char *p = arg;
@@ -857,22 +847,34 @@ parse_cflags (Package *pkg, const char *str, const char *path)
           while (*p && isspace ((guchar)*p))
             ++p;
 
-          pkg->I_cflags = g_slist_prepend (pkg->I_cflags,
-                                           g_strconcat ("-I", p, NULL));
+          flag->type = CFLAGS_I;
+          flag->arg = g_strconcat ("-I", p, NULL);
+          pkg->cflags = g_list_prepend (pkg->cflags, flag);
+        }
+      else if ((strcmp ("-idirafter", arg) == 0 ||
+                strcmp ("-isystem", arg) == 0) &&
+               i+1 < argc)
+        {
+          char *option, *tmp;
 
-        } else {
-          if (*arg != '\0')
-            pkg->other_cflags = g_slist_prepend (pkg->other_cflags,
-                                                 g_strdup (arg));
-	  if (strcmp("-idirafter", arg) == 0) {
-	      char *n;
-
-	      tmp = trim_string(argv[++i]);
-	      n = strdup_escape_shell(tmp);
-	      pkg->other_cflags = g_slist_prepend(pkg->other_cflags, n);
-	      g_free(tmp);
-	  }
-      }
+          tmp = trim_string (argv[i+1]);
+          option = strdup_escape_shell (tmp);
+          flag->type = CFLAGS_OTHER;
+          flag->arg = g_strconcat (arg, " ", option, NULL);
+          pkg->cflags = g_list_prepend (pkg->cflags, flag);
+          i++;
+          g_free (option);
+          g_free (tmp);
+        }
+      else if (*arg != '\0')
+        {
+          flag->type = CFLAGS_OTHER;
+          flag->arg = g_strdup (arg);
+          pkg->cflags = g_list_prepend (pkg->cflags, flag);
+        }
+      else
+        /* flag wasn't used */
+        g_free (flag);
 
       g_free (arg);
       
@@ -889,31 +891,15 @@ parse_url (Package *pkg, const char *str, const char *path)
   if (pkg->url != NULL)
     {
       verbose_error ("URL field occurs twice in '%s'\n", path);
-
-      exit (1);
+      if (parse_strict)
+        exit (1);
+      else
+        return;
     }
 
   pkg->url = trim_and_sub (pkg, str, path);
 }
 
-#ifdef G_OS_WIN32
-static char *orig_prefix = NULL;
-
-static int
-pathnamecmp (const char *a,
-	     const char *b)
-{
-  while (*a && *b &&
-	 ((G_IS_DIR_SEPARATOR (*a) && G_IS_DIR_SEPARATOR (*b)) ||
-	  g_ascii_toupper (*a) == g_ascii_toupper (*b)))
-    {
-      a++;
-      b++;
-    }
-  return g_ascii_toupper (*a) - g_ascii_toupper (*b);
-}
-#endif
-
 static void
 parse_line (Package *pkg, const char *untrimmed, const char *path,
 	    gboolean ignore_requires, gboolean ignore_private_libs,
@@ -1004,41 +990,31 @@ parse_line (Package *pkg, const char *untrimmed, const char *path,
       ++p;
       while (*p && isspace ((guchar)*p))
         ++p;
-      
-      if (pkg->vars == NULL)
-        pkg->vars = g_hash_table_new (g_str_hash, g_str_equal);
 
-#ifdef G_OS_WIN32
-      if (!dont_define_prefix && strcmp (tag, prefix_variable) == 0)
+      if (define_prefix && strcmp (tag, prefix_variable) == 0)
 	{
 	  /* This is the prefix variable. Try to guesstimate a value for it
 	   * for this package from the location of the .pc file.
 	   */
+          gchar *base;
+          gboolean is_pkgconfigdir;
 
-	  gchar *prefix = pkg->pcfiledir;
-	  const int prefix_len = strlen (prefix);
-	  const char *const lib_pkgconfig = "\\lib\\pkgconfig";
-	  const char *const share_pkgconfig = "\\share\\pkgconfig";
-	  const int lib_pkgconfig_len = strlen (lib_pkgconfig);
-	  const int share_pkgconfig_len = strlen (share_pkgconfig);
-
-	  if ((strlen (prefix) > lib_pkgconfig_len &&
-	       pathnamecmp (prefix + prefix_len - lib_pkgconfig_len, lib_pkgconfig) == 0) ||
-	      (strlen (prefix) > share_pkgconfig_len &&
-	       pathnamecmp (prefix + prefix_len - share_pkgconfig_len, share_pkgconfig) == 0))
-	    {
-	      /* It ends in lib\pkgconfig or share\pkgconfig. Good. */
+          base = g_path_get_basename (pkg->pcfiledir);
+          is_pkgconfigdir = g_ascii_strcasecmp (base, "pkgconfig") == 0;
+          g_free (base);
+          if (is_pkgconfigdir)
+            {
+              /* It ends in pkgconfig. Good. */
+              gchar *q;
+              gchar *prefix;
 	      
-	      gchar *q;
-	      
-	      orig_prefix = g_strdup (p);
+              /* Keep track of the original prefix value. */
+              pkg->orig_prefix = g_strdup (p);
 
-	      prefix = g_strdup (prefix);
-	      if (strlen (prefix) > lib_pkgconfig_len &&
-		  pathnamecmp (prefix + prefix_len - lib_pkgconfig_len, lib_pkgconfig) == 0)
-		prefix[prefix_len - lib_pkgconfig_len] = '\0';
-	      else
-		prefix[prefix_len - share_pkgconfig_len] = '\0';
+              /* Get grandparent directory for new prefix. */
+              q = g_path_get_dirname (pkg->pcfiledir);
+              prefix = g_path_get_dirname (q);
+              g_free (q);
 	      
 	      /* Turn backslashes into slashes or
 	       * g_shell_parse_argv() will eat them when ${prefix}
@@ -1051,6 +1027,14 @@ parse_line (Package *pkg, const char *untrimmed, const char *path,
 		    *q = '/';
 		  q++;
 		}
+
+	      /* Now escape the special characters so that there's no danger
+	       * of arguments that include the prefix getting split.
+	       */
+	      q = prefix;
+	      prefix = strdup_escape_shell (prefix);
+	      g_free (q);
+
 	      varname = g_strdup (tag);
 	      debug_spew (" Variable declaration, '%s' overridden with '%s'\n",
 			  tag, prefix);
@@ -1058,24 +1042,26 @@ parse_line (Package *pkg, const char *untrimmed, const char *path,
 	      goto cleanup;
 	    }
 	}
-      else if (!dont_define_prefix &&
-	       orig_prefix != NULL &&
-	       strncmp (p, orig_prefix, strlen (orig_prefix)) == 0 &&
-	       G_IS_DIR_SEPARATOR (p[strlen (orig_prefix)]))
+      else if (define_prefix &&
+	       pkg->orig_prefix != NULL &&
+	       strncmp (p, pkg->orig_prefix, strlen (pkg->orig_prefix)) == 0 &&
+	       G_IS_DIR_SEPARATOR (p[strlen (pkg->orig_prefix)]))
 	{
 	  char *oldstr = str;
 
-	  p = str = g_strconcat (g_hash_table_lookup (pkg->vars, prefix_variable), p + strlen (orig_prefix), NULL);
+	  p = str = g_strconcat (g_hash_table_lookup (pkg->vars, prefix_variable),
+				 p + strlen (pkg->orig_prefix), NULL);
 	  g_free (oldstr);
 	}
-#endif
 
       if (g_hash_table_lookup (pkg->vars, tag))
         {
           verbose_error ("Duplicate definition of variable '%s' in '%s'\n",
                          tag, path);
-
-          exit (1);
+          if (parse_strict)
+            exit (1);
+          else
+            goto cleanup;
         }
 
       varname = g_strdup (tag);
@@ -1125,7 +1111,13 @@ parse_package_file (const char *path, gboolean ignore_requires,
       debug_spew ("No pcfiledir determined for package\n");
       pkg->pcfiledir = g_strdup ("???????");
     }
-  
+
+  if (pkg->vars == NULL)
+    pkg->vars = g_hash_table_new (g_str_hash, g_str_equal);
+
+  /* Variable storing directory of pc file */
+  g_hash_table_insert (pkg->vars, "pcfiledir", pkg->pcfiledir);
+
   str = g_string_new ("");
 
   while (read_one_line (f, str))
@@ -1144,20 +1136,8 @@ parse_package_file (const char *path, gboolean ignore_requires,
   g_string_free (str, TRUE);
   fclose(f);
 
-  /* make ->requires_private include a copy of the public requires too */
-  pkg->requires_private = g_slist_concat(g_slist_copy (pkg->requires),
-					 pkg->requires_private);
-  
-  pkg->requires = g_slist_reverse (pkg->requires);
-  
-  pkg->requires_private = g_slist_reverse (pkg->requires_private);
-
-  pkg->I_cflags = g_slist_reverse (pkg->I_cflags);
-  pkg->other_cflags = g_slist_reverse (pkg->other_cflags);
-
-  pkg->l_libs = g_slist_reverse (pkg->l_libs);
-  pkg->L_libs = g_slist_reverse (pkg->L_libs);
-  pkg->other_libs = g_slist_reverse (pkg->other_libs);
+  pkg->cflags = g_list_reverse (pkg->cflags);
+  pkg->libs = g_list_reverse (pkg->libs);
   
   return pkg;
 }
