@@ -24,6 +24,7 @@
 
 #include "pkg.h"
 #include "parse.h"
+#include "rpmvercmp.h"
 
 #ifdef HAVE_MALLOC_H
 # include <malloc.h>
@@ -126,7 +127,7 @@ name_ends_in_uninstalled (const char *str)
  * locations, ignoring duplicates
  */
 static void
-scan_dir (const char *dirname)
+scan_dir (char *dirname)
 {
   DIR *dir;
   struct dirent *dent;
@@ -159,16 +160,17 @@ scan_dir (const char *dirname)
 #endif
   dir = opendir (dirname_copy);
   g_free (dirname_copy);
+
+  scanned_dir_count += 1;
+
   if (!dir)
     {
-      debug_spew ("Cannot open directory '%s' in package search path: %s\n",
-                  dirname, g_strerror (errno));
+      debug_spew ("Cannot open directory #%i '%s' in package search path: %s\n",
+                  scanned_dir_count, dirname, g_strerror (errno));
       return;
     }
 
-  debug_spew ("Scanning directory '%s'\n", dirname);
-
-  scanned_dir_count += 1;
+  debug_spew ("Scanning directory #%i '%s'\n", scanned_dir_count, dirname);
   
   while ((dent = readdir (dir)))
     {
@@ -176,7 +178,7 @@ scan_dir (const char *dirname)
 
       if (ends_in_dotpc (dent->d_name))
         {
-          char *pkgname = g_malloc (len - 2);
+          char *pkgname = g_malloc (len - EXT_LEN + 1);
 
           debug_spew ("File '%s' appears to be a .pc file\n", dent->d_name);
           
@@ -220,6 +222,7 @@ static Package *
 add_virtual_pkgconfig_package (void)
 {
   Package *pkg = NULL;
+  gchar *path_quoted;
 
   pkg = g_new0 (Package, 1);
 
@@ -232,7 +235,8 @@ add_virtual_pkgconfig_package (void)
 
   if (pkg->vars == NULL)
     pkg->vars = g_hash_table_new (g_str_hash, g_str_equal);
-  g_hash_table_insert (pkg->vars, "pc_path", pkg_config_pc_path);
+  path_quoted = g_shell_quote (pkg_config_pc_path);
+  g_hash_table_insert (pkg->vars, "pc_path", path_quoted);
 
   debug_spew ("Adding virtual 'pkg-config' package to list of known packages\n");
   g_hash_table_insert (packages, pkg->key, pkg);
@@ -916,7 +920,6 @@ verify_package (Package *pkg)
             }
           system_dir_iter = system_dir_iter->next;
         }
-      iter = iter->next;
     }
   g_list_free (system_directories);
 
@@ -988,6 +991,10 @@ packages_get_flags (GList *pkgs, FlagType flags)
       g_free (cur);
     }
 
+  /* Strip trailing space. */
+  if (str->len > 0 && str->str[str->len - 1] == ' ')
+    g_string_truncate (str, str->len - 1);
+
   debug_spew ("returning flags string \"%s\"\n", str->str);
   return g_string_free (str, FALSE);
 }
@@ -1023,10 +1030,6 @@ package_get_var (Package *pkg,
   if (varval == NULL && pkg->vars)
     varval = g_strdup (g_hash_table_lookup (pkg->vars, var));
 
-  /* Magic "pcfiledir" variable */
-  if (varval == NULL && pkg->pcfiledir && strcmp (var, "pcfiledir") == 0)
-    varval = g_strdup (pkg->pcfiledir);
-
   return varval;
 }
 
@@ -1037,6 +1040,7 @@ packages_get_var (GList     *pkgs,
   GList *tmp;
   GString *str;
   char *retval;
+  GError *error = NULL;
   
   str = g_string_new ("");
   
@@ -1044,14 +1048,25 @@ packages_get_var (GList     *pkgs,
   while (tmp != NULL)
     {
       Package *pkg = tmp->data;
-      char *var;
+      char *var, *unquoted_var;
 
       var = package_get_var (pkg, varname);
-      
       if (var)
         {
-          g_string_append (str, var);
-          g_string_append_c (str, ' ');                
+          unquoted_var = g_shell_unquote (var, &error);
+          if (unquoted_var != NULL)
+            {
+              g_string_append (str, unquoted_var);
+              g_string_append_c (str, ' ');
+              g_free (unquoted_var);
+            }
+          else
+            {
+              verbose_error ("Couldn't unquote value of \"%s\": %s\n",
+                             varname, error ? error->message : "unknown");
+              g_clear_error (&error);
+            }
+
           g_free (var);
         }
 
@@ -1067,109 +1082,6 @@ packages_get_var (GList     *pkgs,
   return retval;
 }
 
-
-
-/* Stolen verbatim from rpm/lib/misc.c 
-   RPM is Copyright (c) 1998 by Red Hat Software, Inc.,
-   and may be distributed under the terms of the GPL and LGPL.
-*/
-/* compare alpha and numeric segments of two versions */
-/* return 1: a is newer than b */
-/*        0: a and b are the same version */
-/*       -1: b is newer than a */
-static int rpmvercmp(const char * a, const char * b) {
-    char oldch1, oldch2;
-    char * str1, * str2;
-    char * one, * two;
-    int rc;
-    int isnum;
-    
-    /* easy comparison to see if versions are identical */
-    if (!strcmp(a, b)) return 0;
-
-    str1 = g_alloca(strlen(a) + 1);
-    str2 = g_alloca(strlen(b) + 1);
-
-    strcpy(str1, a);
-    strcpy(str2, b);
-
-    one = str1;
-    two = str2;
-
-    /* loop through each version segment of str1 and str2 and compare them */
-    while (*one && *two) {
-	while (*one && !isalnum((guchar)*one)) one++;
-	while (*two && !isalnum((guchar)*two)) two++;
-
-	/* If we ran to the end of either, we are finished with the loop */
-	if (!(*one && *two)) break;
-
-	str1 = one;
-	str2 = two;
-
-	/* grab first completely alpha or completely numeric segment */
-	/* leave one and two pointing to the start of the alpha or numeric */
-	/* segment and walk str1 and str2 to end of segment */
-	if (isdigit((guchar)*str1)) {
-	    while (*str1 && isdigit((guchar)*str1)) str1++;
-	    while (*str2 && isdigit((guchar)*str2)) str2++;
-	    isnum = 1;
-	} else {
-	    while (*str1 && isalpha((guchar)*str1)) str1++;
-	    while (*str2 && isalpha((guchar)*str2)) str2++;
-	    isnum = 0;
-	}
-		
-	/* save character at the end of the alpha or numeric segment */
-	/* so that they can be restored after the comparison */
-	oldch1 = *str1;
-	*str1 = '\0';
-	oldch2 = *str2;
-	*str2 = '\0';
-
-	/* take care of the case where the two version segments are */
-	/* different types: one numeric and one alpha */
-	if (one == str1) return -1;	/* arbitrary */
-	/* XXX See patch #60884 (and details) from bugzilla #50977. */
-	if (two == str2) return (isnum ? 1 : -1);
-
-	if (isnum) {
-	    /* this used to be done by converting the digit segments */
-	    /* to ints using atoi() - it's changed because long  */
-	    /* digit segments can overflow an int - this should fix that. */
-	  
-	    /* throw away any leading zeros - it's a number, right? */
-	    while (*one == '0') one++;
-	    while (*two == '0') two++;
-
-	    /* whichever number has more digits wins */
-	    if (strlen(one) > strlen(two)) return 1;
-	    if (strlen(two) > strlen(one)) return -1;
-	}
-
-	/* strcmp will return which one is greater - even if the two */
-	/* segments are alpha or if they are numeric.  don't return  */
-	/* if they are equal because there might be more segments to */
-	/* compare */
-	rc = strcmp(one, two);
-	if (rc) return rc;
-	
-	/* restore character that was replaced by null above */
-	*str1 = oldch1;
-	one = str1;
-	*str2 = oldch2;
-	two = str2;
-    }
-
-    /* this catches the case where all numeric and alpha segments have */
-    /* compared identically but the segment sepparating characters were */
-    /* different */
-    if ((!*one) && (!*two)) return 0;
-
-    /* whichever version still has characters left over wins */
-    if (!*one) return -1; else return 1;
-}
-
 int
 compare_versions (const char * a, const char *b)
 {
